from .base import BaseLightfuzz


class ssti(BaseLightfuzz):
    """
    Detects server-side template injection vulnerabilities.

    Techniques:

    * Arithmetic Evaluation:
       - Injects encoded and unencoded multiplication expressions to detect evaluation
    """

    friendly_name = "Server-side Template Injection"

    async def fuzz(self):
        cookies = self.event.data.get("assigned_cookies", {})
        # These are common SSTI payloads, each attempting to trigger an integer multiplication which would produce an expected value
        ssti_probes = [
            "<%25%3d%201337*1337%20%25>",
            "<%= 1337*1337 %>",
            "${1337*1337}",
            "%24%7b1337*1337%7d",
            "1,787{{z}},569",
        ]
        for probe_value in ssti_probes:
            r = await self.standard_probe(
                self.event.data["type"], cookies, probe_value, allow_redirects=True, skip_urlencoding=True
            )

            # look for the expected value in the response
            if r and ("1787569" in r.text or "1,787,569" in r.text):
                self.results.append(
                    {
                        "type": "FINDING",
                        "description": f"POSSIBLE Server-side Template Injection. {self.metadata()} Detection Method: [Integer Multiplication] Payload: [{probe_value}]",
                    }
                )
                break



# dnscaa.py
#
# Checks for and parses CAA DNS TXT records for IODEF reporting destination email addresses and/or URL's.
#
# NOTE: when the target domain is initially resolved basic "dns_name_extraction_regex" matched targets will be extracted so we do not perform that again here.
#
# Example CAA records,
#   0 iodef "mailto:dnsadmin@example.com"
#   0 iodef "mailto:contact_pki@example.com"
#   0 iodef "mailto:ipladmin@example.com"
#   0 iodef "https://example.com/caa"
#   0 iodef "https://203.0.113.1/caa" <<< unlikely but possible?
#   0 iodef "https://[2001:db8::1]/caa" <<< unlikely but possible?
#
# We simply extract any URL's as URL_UNVERIFIED, no further activity against URL's is performed by this module in order to remain passive.
#
# Other modules which respond to URL_UNVERIFIED events may do so if you have configured bbot appropriately.
#
# The domain/IP portion of any URL_UNVERIFIED's should be extracted by the various internal modules.
#

from bbot.modules.base import BaseModule

import re

from bbot.core.helpers.regexes import dns_name_extraction_regex, email_regex, url_regexes

# Handle '0 iodef "mailto:support@hcaptcha.com"'
# Handle '1 iodef "https://some.host.tld/caa;"'
# Handle '0 issue "pki.goog; cansignhttpexchanges=yes; somethingelse=1"'
# Handle '1 issue ";"' == explicit denial for any wildcard issuance.
# Handle '128 issuewild "comodoca.com"'
# Handle '128 issuewild ";"' == explicit denial for any wildcard issuance.
_caa_regex = r"^(?P<flags>[0-9]+) +(?P<property>\w+) +\"(?P<text>[^;\"]*);* *(?P<extensions>[^\"]*)\"$"
caa_regex = re.compile(_caa_regex)

_caa_extensions_kvp_regex = r"(?P<k>\w+)=(?P<v>[^;]+)"
caa_extensions_kvp_regex = re.compile(_caa_extensions_kvp_regex)


class dnscaa(BaseModule):
    watched_events = ["DNS_NAME"]
    produced_events = ["DNS_NAME", "EMAIL_ADDRESS", "URL_UNVERIFIED"]
    flags = ["subdomain-enum", "email-enum", "passive", "safe"]
    meta = {"description": "Check for CAA records", "author": "@colin-stubbs", "created_date": "2024-05-26"}
    options = {
        "in_scope_only": True,
        "dns_names": True,
        "emails": True,
        "urls": True,
    }
    options_desc = {
        "in_scope_only": "Only check in-scope domains",
        "dns_names": "emit DNS_NAME events",
        "emails": "emit EMAIL_ADDRESS events",
        "urls": "emit URL_UNVERIFIED events",
    }
    # accept DNS_NAMEs out to 2 hops if in_scope_only is False
    scope_distance_modifier = 2

    async def setup(self):
        self.in_scope_only = self.config.get("in_scope_only", True)
        self._dns_names = self.config.get("dns_names", True)
        self._emails = self.config.get("emails", True)
        self._urls = self.config.get("urls", True)
        return await super().setup()

    async def filter_event(self, event):
        if "_wildcard" in str(event.host).split("."):
            return False, "event is wildcard"

        # scope filtering
        if event.scope_distance > 0 and self.in_scope_only:
            return False, "event is not in scope"

        return True

    async def handle_event(self, event):
        tags = ["caa-record"]

        r = await self.helpers.resolve_raw(event.host, type="caa")

        if r:
            raw_results, errors = r

            for answer in raw_results:
                s = answer.to_text().strip().replace('" "', "")

                # validate CAA record vi regex so that we can determine what to do with it.
                caa_match = caa_regex.search(s)

                if caa_match and caa_match.group("flags") and caa_match.group("property") and caa_match.group("text"):
                    # it's legit.
                    if caa_match.group("property").lower() == "iodef":
                        if self._emails:
                            for match in email_regex.finditer(caa_match.group("text")):
                                start, end = match.span()
                                email = caa_match.group("text")[start:end]

                                await self.emit_event(email, "EMAIL_ADDRESS", tags=tags, parent=event)

                        if self._urls:
                            for url_regex in url_regexes:
                                for match in url_regex.finditer(caa_match.group("text")):
                                    start, end = match.span()
                                    url = caa_match.group("text")[start:end].strip('"').strip()

                                    await self.emit_event(url, "URL_UNVERIFIED", tags=tags, parent=event)

                    elif caa_match.group("property").lower().startswith("issue"):
                        if self._dns_names:
                            for match in dns_name_extraction_regex.finditer(caa_match.group("text")):
                                start, end = match.span()
                                name = caa_match.group("text")[start:end]

                                await self.emit_event(name, "DNS_NAME", tags=tags, parent=event)


# EOF



from .base import ModuleTestBase


class TestHackertarget(ModuleTestBase):
    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.hackertarget.com/hostsearch/?q=blacklanternsecurity.com",
            text="asdf.blacklanternsecurity.com\nzzzz.blacklanternsecurity.com",
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"
        assert any(e.data == "zzzz.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from .base import ModuleTestBase


class TestCertspotter(ModuleTestBase):
    async def setup_after_prep(self, module_test):
        module_test.module.abort_if = lambda e: False
        for t in self.targets:
            module_test.httpx_mock.add_response(
                url="https://api.certspotter.com/v1/issuances?domain=blacklanternsecurity.com&include_subdomains=true&expand=dns_names",
                json=[{"dns_names": ["*.asdf.blacklanternsecurity.com"]}],
            )

    def check(self, module_test, events):
        assert any(e.data == "asdf.blacklanternsecurity.com" for e in events), "Failed to detect subdomain"



from bbot.errors import HttpCompareError
from .base import BaseLightfuzz

import urllib.parse


class cmdi(BaseLightfuzz):
    """
    Detects command injection vulnerabilities.

    Techniques:

    * Echo Canary Detection:
       - Injects command delimiters (;, &&, ||, &, |) along with an echo command
       - Checks if the echoed canary appears in the response without the "echo" itself
       - Uses a false positive probe to validate findings

    * Blind Command Injection:
       - Injects nslookup commands with unique subdomain tags
       - Detects command execution through DNS resolution via Interactsh
    """

    friendly_name = "Command Injection"
    uses_interactsh = True

    async def fuzz(self):
        cookies = self.event.data.get(
            "assigned_cookies", {}
        )  # Retrieve assigned cookies from WEB_PARAMETER event data, if present
        probe_value = self.incoming_probe_value()

        canary = self.lightfuzz.helpers.rand_string(10, numeric_only=True)
        http_compare = self.compare_baseline(
            self.event.data["type"], probe_value, cookies
        )  # Initialize the http_compare object and establish a baseline HTTP response

        cmdi_probe_strings = [
            "AAAA",  # False positive probe
            ";",
            "&&",
            "||",
            "&",
            "|",
        ]

        positive_detections = []
        for p in cmdi_probe_strings:
            try:
                # add "echo" to the cmdi probe value to construct the command to be executed
                echo_probe = f"{probe_value}{p} echo {canary} {p}"
                # we have to handle our own URL-encoding here, because our payloads include the & character
                if self.event.data["type"] == "GETPARAM":
                    echo_probe = urllib.parse.quote(echo_probe.encode(), safe="")

                # send cmdi probe and compare with baseline response
                cmdi_probe = await self.compare_probe(
                    http_compare, self.event.data["type"], echo_probe, cookies, skip_urlencoding=True
                )

                # ensure we received an HTTP response
                if cmdi_probe[3]:
                    # check if the canary is in the response and the word "echo" is NOT in the response text, ruling out mere reflection of the entire probe value without execution
                    if canary in cmdi_probe[3].text and "echo" not in cmdi_probe[3].text:
                        self.debug(f"canary [{canary}] found in response when sending probe [{p}]")
                        if p == "AAAA":  # Handle detection false positive probe
                            self.warning(
                                f"False Postive Probe appears to have been triggered for {self.event.data['url']}, aborting remaining detection"
                            )
                            return
                        positive_detections.append(p)  # Add detected probes to positive detections
            except HttpCompareError as e:
                self.debug(e)
                continue
        if len(positive_detections) > 0:
            self.results.append(
                {
                    "type": "FINDING",
                    "description": f"POSSIBLE OS Command Injection. {self.metadata()} Detection Method: [echo canary] CMD Probe Delimeters: [{' '.join(positive_detections)}]",
                }
            )

        # Blind OS Command Injection
        if self.lightfuzz.interactsh_instance:
            self.lightfuzz.event_dict[self.event.data["url"]] = self.event  # Store the event associated with the URL
            for p in cmdi_probe_strings:
                # generate a random subdomain tag and associate it with the event, type, name, and probe
                subdomain_tag = self.lightfuzz.helpers.rand_string(4, digits=False)
                self.lightfuzz.interactsh_subdomain_tags[subdomain_tag] = {
                    "event": self.event,
                    "type": self.event.data["type"],
                    "name": self.event.data["name"],
                    "probe": p,
                }
                # payload is an nslookup command that includes the interactsh domain prepended the previously generated subdomain tag
                interactsh_probe = f"{p} nslookup {subdomain_tag}.{self.lightfuzz.interactsh_domain} {p}"
                # we have to handle our own URL-encoding here, because our payloads include the & character
                if self.event.data["type"] == "GETPARAM":
                    interactsh_probe = urllib.parse.quote(interactsh_probe.encode(), safe="")
                # we send the probe here, and any positive detections are processed in the interactsh_callback defined in lightfuzz.py
                await self.standard_probe(
                    self.event.data["type"],
                    cookies,
                    f"{probe_value}{interactsh_probe}",
                    timeout=15,
                    skip_urlencoding=True,
                )



from .base import ModuleTestBase


class TestSocial(ModuleTestBase):
    targets = ["http://127.0.0.1:8888"]
    modules_overrides = ["httpx", "excavate", "social"]

    async def setup_after_prep(self, module_test):
        expect_args = {"method": "GET", "uri": "/"}
        respond_args = {
            "response_data": """
            <html>
                <a href="https://discord.gg/asdf"/><a href="https://github.com/blacklanternsecurity/bbot"/>
                <a href="https://hub.docker.com/r/blacklanternsecurity"/>
                <a href="https://hub.docker.com/r/blacklanternsecurity/bbot"/>
                <a href="https://hub.docker.com/r/blacklanternSECURITY/bbot"/>
                <a href="https://www.postman.com/blacklanternsecurity/bbot"/>
            </html>
            """
        }
        module_test.set_expect_requests(expect_args=expect_args, respond_args=respond_args)

    def check(self, module_test, events):
        assert 4 == len([e for e in events if e.type == "SOCIAL"])
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL" and e.data["platform"] == "discord" and e.data["profile_name"] == "asdf"
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "docker"
                and e.data["profile_name"] == "blacklanternsecurity"
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "github"
                and e.data["profile_name"] == "blacklanternsecurity"
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "SOCIAL"
                and e.data["platform"] == "postman"
                and e.data["profile_name"] == "blacklanternsecurity"
            ]
        )



from .base import ModuleTestBase


class TestPortscan(ModuleTestBase):
    targets = [
        "www.evilcorp.com",
        "evilcorp.com",
        "8.8.8.8/32",
        "8.8.8.8/24",
        "8.8.4.4",
        "asdf.evilcorp.net",
        "8.8.4.4/24",
    ]
    scan_name = "test_portscan"
    config_overrides = {"modules": {"portscan": {"ports": "443", "wait": 1}}, "dns": {"minimal": False}}

    masscan_output_1 = """{   "ip": "8.8.8.8",   "timestamp": "1680197558", "ports": [ {"port": 443, "proto": "tcp", "status": "open", "reason": "syn-ack", "ttl": 54} ] }"""
    masscan_output_2 = """{   "ip": "8.8.4.5",   "timestamp": "1680197558", "ports": [ {"port": 80, "proto": "tcp", "status": "open", "reason": "syn-ack", "ttl": 54} ] }"""
    masscan_output_3 = """{   "ip": "8.8.4.6",   "timestamp": "1680197558", "ports": [ {"port": 631, "proto": "tcp", "status": "open", "reason": "syn-ack", "ttl": 54} ] }"""

    masscan_output_ping = """{   "ip": "8.8.8.8",   "timestamp": "1719862594", "ports": [ {"port": 0, "proto": "icmp", "status": "open", "reason": "none", "ttl": 54} ] }"""

    async def setup_after_prep(self, module_test):
        from bbot.modules.base import BaseModule

        class DummyModule(BaseModule):
            _name = "dummy_module"
            watched_events = ["*"]

            async def handle_event(self, event):
                if event.type == "DNS_NAME":
                    if "dummy" not in event.host:
                        await self.emit_event(f"dummy.{event.data}", "DNS_NAME", parent=event)

        module_test.scan.modules["dummy_module"] = DummyModule(module_test.scan)

        await module_test.mock_dns(
            {
                "www.evilcorp.com": {"A": ["8.8.8.8"]},
                "evilcorp.com": {"A": ["8.8.8.8"]},
                "asdf.evilcorp.net": {"A": ["8.8.4.5"]},
                "dummy.asdf.evilcorp.net": {"A": ["8.8.4.5"]},
                "dummy.evilcorp.com": {"A": ["8.8.4.6"]},
                "dummy.www.evilcorp.com": {"A": ["8.8.4.4"]},
            }
        )

        self.syn_scanned = []
        self.ping_scanned = []
        self.syn_runs = 0
        self.ping_runs = 0

        async def run_masscan(command, *args, **kwargs):
            if "masscan" in command[:2]:
                targets = open(command[11]).read().splitlines()
                yield "["
                if "--ping" in command:
                    self.ping_runs += 1
                    self.ping_scanned += targets
                    yield self.masscan_output_ping
                else:
                    self.syn_runs += 1
                    self.syn_scanned += targets
                    if "8.8.8.0/24" in targets or "8.8.8.8/32" in targets:
                        yield self.masscan_output_1
                    if "8.8.4.0/24" in targets:
                        yield self.masscan_output_2
                        yield self.masscan_output_3
                yield "]"
            else:
                async for l in module_test.scan.helpers.run_live(command, *args, **kwargs):
                    yield l

        module_test.monkeypatch.setattr(module_test.scan.helpers, "run_live", run_masscan)

    def check(self, module_test, events):
        assert set(self.syn_scanned) == {"8.8.8.0/24", "8.8.4.0/24"}
        assert set(self.ping_scanned) == set()
        assert self.syn_runs >= 1
        assert self.ping_runs == 0
        assert 1 == len(
            [e for e in events if e.type == "DNS_NAME" and e.data == "evilcorp.com" and str(e.module) == "TARGET"]
        )
        assert 1 == len(
            [e for e in events if e.type == "DNS_NAME" and e.data == "www.evilcorp.com" and str(e.module) == "TARGET"]
        )
        assert 1 == len(
            [e for e in events if e.type == "DNS_NAME" and e.data == "asdf.evilcorp.net" and str(e.module) == "TARGET"]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME" and e.data == "dummy.evilcorp.com" and str(e.module) == "dummy_module"
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME" and e.data == "dummy.www.evilcorp.com" and str(e.module) == "dummy_module"
            ]
        )
        assert 1 == len(
            [
                e
                for e in events
                if e.type == "DNS_NAME" and e.data == "dummy.asdf.evilcorp.net" and str(e.module) == "dummy_module"
            ]
        )
        # the reason these numbers aren't exactly predictable is because we can't predict which one arrives first
        # to the portscan module. Sometimes, one that would normally be deduped is force-emitted because it led to a new open port.
        assert 2 <= len([e for e in events if e.type == "IP_ADDRESS" and e.data == "8.8.8.8"]) <= 4
        assert 2 <= len([e for e in events if e.type == "IP_ADDRESS" and e.data == "8.8.4.4"]) <= 4
        assert 2 <= len([e for e in events if e.type == "IP_ADDRESS" and e.data == "8.8.4.5"]) <= 4
        assert 2 <= len([e for e in events if e.type == "IP_ADDRESS" and e.data == "8.8.4.6"]) <= 4
        assert 1 == len([e for e in events if e.type == "OPEN_TCP_PORT" and e.data == "8.8.8.8:443"])
        assert 1 == len([e for e in events if e.type == "OPEN_TCP_PORT" and e.data == "8.8.4.5:80"])
        assert 1 == len([e for e in events if e.type == "OPEN_TCP_PORT" and e.data == "8.8.4.6:631"])
        assert 1 == len([e for e in events if e.type == "OPEN_TCP_PORT" and e.data == "evilcorp.com:443"])
        assert 1 == len([e for e in events if e.type == "OPEN_TCP_PORT" and e.data == "www.evilcorp.com:443"])
        assert 1 == len([e for e in events if e.type == "OPEN_TCP_PORT" and e.data == "asdf.evilcorp.net:80"])
        assert 1 == len([e for e in events if e.type == "OPEN_TCP_PORT" and e.data == "dummy.asdf.evilcorp.net:80"])
        assert 1 == len([e for e in events if e.type == "OPEN_TCP_PORT" and e.data == "dummy.evilcorp.com:631"])
        assert not any(e for e in events if e.type == "OPEN_TCP_PORT" and e.host == "dummy.www.evilcorp.com")


class TestPortscanPingFirst(TestPortscan):
    modules_overrides = {"portscan"}
    config_overrides = {"modules": {"portscan": {"ports": "443", "wait": 1, "ping_first": True}}}

    def check(self, module_test, events):
        assert set(self.syn_scanned) == {"8.8.8.8/32"}
        assert set(self.ping_scanned) == {"8.8.8.0/24", "8.8.4.0/24"}
        assert self.syn_runs == 1
        assert self.ping_runs >= 1
        open_port_events = [e for e in events if e.type == "OPEN_TCP_PORT"]
        assert len(open_port_events) == 3
        assert {e.data for e in open_port_events} == {"8.8.8.8:443", "evilcorp.com:443", "www.evilcorp.com:443"}


class TestPortscanPingOnly(TestPortscan):
    modules_overrides = {"portscan"}
    config_overrides = {"modules": {"portscan": {"ports": "443", "wait": 1, "ping_only": True}}}

    targets = ["8.8.8.8/24", "8.8.4.4/24"]

    def check(self, module_test, events):
        assert set(self.syn_scanned) == set()
        assert set(self.ping_scanned) == {"8.8.8.0/24", "8.8.4.0/24"}
        assert self.syn_runs == 0
        assert self.ping_runs >= 1
        open_port_events = [e for e in events if e.type == "OPEN_TCP_PORT"]
        assert len(open_port_events) == 0
        ip_events = [e for e in events if e.type == "IP_ADDRESS"]
        assert len(ip_events) == 1
        assert {e.data for e in ip_events} == {"8.8.8.8"}



from bbot.modules.templates.github import github
from bbot.modules.templates.subdomain_enum import subdomain_enum


class github_codesearch(github, subdomain_enum):
    watched_events = ["DNS_NAME"]
    produced_events = ["CODE_REPOSITORY", "URL_UNVERIFIED"]
    flags = ["passive", "subdomain-enum", "safe", "code-enum"]
    meta = {
        "description": "Query Github's API for code containing the target domain name",
        "created_date": "2023-12-14",
        "author": "@domwhewell-sage",
        "auth_required": True,
    }
    options = {"api_key": "", "limit": 100}
    options_desc = {"api_key": "Github token", "limit": "Limit code search to this many results"}

    github_raw_url = "https://raw.githubusercontent.com/"

    async def setup(self):
        self.limit = self.config.get("limit", 100)
        return await super().setup()

    async def handle_event(self, event):
        query = self.make_query(event)
        for repo_url, raw_urls in (await self.query(query)).items():
            repo_event = self.make_event({"url": repo_url}, "CODE_REPOSITORY", tags="git", parent=event)
            if repo_event is None:
                continue
            await self.emit_event(
                repo_event,
                context=f'{{module}} searched github.com for "{query}" and found {{event.type}} with matching content at {repo_url}',
            )
            for raw_url in raw_urls:
                url_event = self.make_event(raw_url, "URL_UNVERIFIED", parent=repo_event, tags=["httpx-safe"])
                if not url_event:
                    continue
                await self.emit_event(
                    url_event, context=f'file matching query "{query}" is at {{event.type}}: {raw_url}'
                )

    async def query(self, query):
        repos = {}
        url = f"{self.base_url}/search/code?per_page=100&type=Code&q={self.helpers.quote(query)}&page=" + "{page}"
        agen = self.api_page_iter(url, headers=self.headers, _json=False)
        num_results = 0
        try:
            async for r in agen:
                if r is None:
                    break
                status_code = getattr(r, "status_code", 0)
                if status_code == 429:
                    self.info("Github is rate-limiting us (HTTP status: 429)")
                    break
                if status_code != 200:
                    self.info(f"Unexpected response (HTTP status: {status_code})")
                    break
                try:
                    j = r.json()
                except Exception as e:
                    self.warning(f"Failed to decode JSON for {r.url} (HTTP status: {status_code}): {e}")
                    break
                items = j.get("items", [])
                if not items:
                    break
                for item in items:
                    html_url = item.get("html_url", "")
                    raw_url = self.raw_url(html_url)
                    repo_url = item.get("repository", {}).get("html_url", "")
                    if raw_url and repo_url:
                        try:
                            repos[repo_url].append(raw_url)
                        except KeyError:
                            repos[repo_url] = [raw_url]
                        num_results += 1
                        if num_results >= self.limit:
                            break
                if num_results >= self.limit:
                    break
        finally:
            await agen.aclose()
        return repos

    def raw_url(self, url):
        return url.replace("https://github.com/", self.github_raw_url).replace("/blob/", "/")



import sys
from xml.dom import minidom
from datetime import datetime
from xml.etree.ElementTree import Element, SubElement, tostring

from bbot import __version__
from bbot.modules.output.base import BaseOutputModule


class NmapHost:
    __slots__ = ["hostnames", "open_ports"]

    def __init__(self):
        self.hostnames = set()
        # a dict of {port: {protocol: banner}}
        self.open_ports = dict()


class Nmap_XML(BaseOutputModule):
    watched_events = ["OPEN_TCP_PORT", "DNS_NAME", "IP_ADDRESS", "PROTOCOL", "HTTP_RESPONSE"]
    meta = {"description": "Output to Nmap XML", "created_date": "2024-11-16", "author": "@TheTechromancer"}
    output_filename = "output.nmap.xml"
    in_scope_only = True

    async def setup(self):
        self.hosts = {}
        self._prep_output_dir(self.output_filename)
        return True

    async def handle_event(self, event):
        event_host = event.host

        # we always record by IP
        ips = []
        for ip in event.resolved_hosts:
            try:
                ips.append(self.helpers.make_ip_type(ip))
            except ValueError:
                continue
        if not ips and self.helpers.is_ip(event_host):
            ips = [event_host]

        for ip in ips:
            try:
                nmap_host = self.hosts[ip]
            except KeyError:
                nmap_host = NmapHost()
                self.hosts[ip] = nmap_host

            event_port = getattr(event, "port", None)
            if event.type == "OPEN_TCP_PORT":
                if event_port not in nmap_host.open_ports:
                    nmap_host.open_ports[event.port] = {}
            elif event.type in ("PROTOCOL", "HTTP_RESPONSE"):
                if event_port is not None:
                    try:
                        existing_services = nmap_host.open_ports[event.port]
                    except KeyError:
                        existing_services = {}
                        nmap_host.open_ports[event.port] = existing_services
                    if event.type == "PROTOCOL":
                        protocol = event.data["protocol"].lower()
                        banner = event.data.get("banner", None)
                    elif event.type == "HTTP_RESPONSE":
                        protocol = event.parsed_url.scheme.lower()
                        banner = event.http_title
                    if protocol not in existing_services:
                        existing_services[protocol] = banner

            if self.helpers.is_ip(event_host):
                if str(event.module) == "PTR":
                    nmap_host.hostnames.add(event.parent.data)
            else:
                nmap_host.hostnames.add(event_host)

    async def report(self):
        scan_start_time = str(int(self.scan.start_time.timestamp()))
        scan_start_time_str = self.scan.start_time.strftime("%a %b %d %H:%M:%S %Y")
        scan_end_time = datetime.now()
        scan_end_time_str = scan_end_time.strftime("%a %b %d %H:%M:%S %Y")
        scan_end_time_timestamp = str(scan_end_time.timestamp())
        scan_duration = scan_end_time - self.scan.start_time
        num_hosts_up = len(self.hosts)

        # Create the root element
        nmaprun = Element(
            "nmaprun",
            {
                "scanner": "bbot",
                "args": " ".join(sys.argv),
                "start": scan_start_time,
                "startstr": scan_start_time_str,
                "version": str(__version__),
                "xmloutputversion": "1.05",
            },
        )

        ports_scanned = []
        speculate_module = self.scan.modules.get("speculate", None)
        if speculate_module is not None:
            ports_scanned = speculate_module.ports
        portscan_module = self.scan.modules.get("portscan", None)
        if portscan_module is not None:
            ports_scanned = self.helpers.parse_port_string(str(portscan_module.ports))
        num_ports_scanned = len(sorted(ports_scanned))
        ports_scanned = ",".join(str(x) for x in sorted(ports_scanned))

        # Add scaninfo
        SubElement(
            nmaprun,
            "scaninfo",
            {"type": "syn", "protocol": "tcp", "numservices": str(num_ports_scanned), "services": ports_scanned},
        )

        # Add host information
        for ip, nmap_host in self.hosts.items():
            hostnames = sorted(nmap_host.hostnames)
            ports = sorted(nmap_host.open_ports)

            host_elem = SubElement(nmaprun, "host")
            SubElement(host_elem, "status", {"state": "up", "reason": "user-set", "reason_ttl": "0"})
            SubElement(host_elem, "address", {"addr": str(ip), "addrtype": f"ipv{ip.version}"})

            if hostnames:
                hostnames_elem = SubElement(host_elem, "hostnames")
                for hostname in hostnames:
                    SubElement(hostnames_elem, "hostname", {"name": hostname, "type": "user"})

            ports = SubElement(host_elem, "ports")
            for port, protocols in nmap_host.open_ports.items():
                port_elem = SubElement(ports, "port", {"protocol": "tcp", "portid": str(port)})
                SubElement(port_elem, "state", {"state": "open", "reason": "syn-ack", "reason_ttl": "0"})
                # <port protocol="tcp" portid="443"><state state="open" reason="syn-ack" reason_ttl="53"/><service name="http" product="AkamaiGHost" extrainfo="Akamai&apos;s HTTP Acceleration/Mirror service" tunnel="ssl" method="probed" conf="10"/></port>
                for protocol, banner in protocols.items():
                    attrs = {"name": protocol, "method": "probed", "conf": "10"}
                    if banner is not None:
                        attrs["product"] = banner
                        attrs["extrainfo"] = banner
                    SubElement(port_elem, "service", attrs)

        # Add runstats
        runstats = SubElement(nmaprun, "runstats")
        SubElement(
            runstats,
            "finished",
            {
                "time": scan_end_time_timestamp,
                "timestr": scan_end_time_str,
                "summary": f"BBOT done at {scan_end_time_str}; {num_hosts_up} scanned in {scan_duration} seconds",
                "elapsed": str(scan_duration.total_seconds()),
                "exit": "success",
            },
        )
        SubElement(runstats, "hosts", {"up": str(num_hosts_up), "down": "0", "total": str(num_hosts_up)})

        # make backup of the file
        self.helpers.backup_file(self.output_file)

        # Pretty-format the XML
        rough_string = tostring(nmaprun, encoding="utf-8")
        reparsed = minidom.parseString(rough_string)

        # Create a new document with the doctype
        doctype = minidom.DocumentType("nmaprun")
        reparsed.insertBefore(doctype, reparsed.documentElement)

        pretty_xml = reparsed.toprettyxml(indent="  ")

        with open(self.output_file, "w") as f:
            f.write(pretty_xml)
        self.info(f"Saved Nmap XML output to {self.output_file}")



from .base import ModuleTestBase, tempwordlist
import pytest


@pytest.fixture
def mock_medusa_run_process(monkeypatch):
    async def fake_run_process(self, cmd):
        class FakeResult:
            stdout = "ACCOUNT FOUND: [snmp] Host: 127.0.0.1 User: (null) Password: public [ERROR]\n"
            stderr = (
                "ERROR: [snmp.mod] Error processing SNMP response (1).\n"
                "ERROR: [snmp.mod] Community string appears to have only READ access.\n"
            )

        return FakeResult()

    from bbot.modules.base import BaseModule

    monkeypatch.setattr(BaseModule, "run_process", fake_run_process)


@pytest.mark.usefixtures("mock_medusa_run_process")
class TestMedusa(ModuleTestBase):
    targets = ["127.0.0.1"]
    temp_snmp_wordlist = tempwordlist(["public", "private, admin"])
    config_overrides = {
        "modules": {
            "medusa": {
                "snmp_versions": ["2C"],
                "timeout_s": 1,
                "snmp_wordlist": str(temp_snmp_wordlist),
            }
        }
    }

    async def setup_after_prep(self, module_test):
        protocol_data = {"host": str(self.targets[0]), "protocol": "snmp", "port": 161}

        protocol_event = module_test.scan.make_event(
            protocol_data,
            "PROTOCOL",
            parent=module_test.scan.root_event,
        )
        await module_test.module.emit_event(protocol_event)

    def check(self, module_test, events):
        vuln_events = [e for e in events if e.type == "VULNERABILITY"]

        assert len(vuln_events) == 1
        assert "VALID [SNMPV2C] CREDENTIALS FOUND: public [READ]" in vuln_events[0].data["description"]



from .base import ModuleTestBase


class TestHunterio(ModuleTestBase):
    config_overrides = {"modules": {"hunterio": {"api_key": ["asdf", "1234", "4321", "fdsa"]}}}

    async def setup_before_prep(self, module_test):
        module_test.httpx_mock.add_response(
            url="https://api.hunter.io/v2/account?api_key=asdf",
            json={
                "data": {
                    "first_name": "jon",
                    "last_name": "snow",
                    "email": "jon@blacklanternsecurity.notreal",
                    "plan_name": "Starter",
                    "plan_level": 1,
                    "reset_date": "1917-05-23",
                    "team_id": 1234,
                    "calls": {
                        "_deprecation_notice": "Sums the searches and the verifications, giving an imprecise look of the available requests",
                        "used": 999,
                        "available": 2000,
                    },
                    "requests": {
                        "searches": {"used": 998, "available": 1000},
                        "verifications": {"used": 0, "available": 1000},
                    },
                }
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.hunter.io/v2/domain-search?domain=blacklanternsecurity.com&api_key=fdsa&limit=100&offset=0",
            json={
                "data": {
                    "domain": "blacklanternsecurity.com",
                    "disposable": False,
                    "webmail": False,
                    "accept_all": False,
                    "pattern": "{first}",
                    "organization": "Black Lantern Security",
                    "description": None,
                    "twitter": None,
                    "facebook": None,
                    "linkedin": "https://linkedin.com/company/black-lantern-security",
                    "instagram": None,
                    "youtube": None,
                    "technologies": ["jekyll", "nginx"],
                    "country": "US",
                    "state": "CA",
                    "city": "Night City",
                    "postal_code": "12345",
                    "street": "123 Any St",
                    "emails": [
                        {
                            "value": "asdf@blacklanternsecurity.com",
                            "type": "generic",
                            "confidence": 77,
                            "sources": [
                                {
                                    "domain": "blacklanternsecurity.com",
                                    "uri": "http://blacklanternsecurity.com",
                                    "extracted_on": "2021-06-09",
                                    "last_seen_on": "2023-03-21",
                                    "still_on_page": True,
                                }
                            ],
                            "first_name": None,
                            "last_name": None,
                            "position": None,
                            "seniority": None,
                            "department": "support",
                            "linkedin": None,
                            "twitter": None,
                            "phone_number": None,
                            "verification": {"date": None, "status": None},
                        }
                    ],
                    "linked_domains": [],
                },
                "meta": {
                    "results": 1,
                    "limit": 100,
                    "offset": 0,
                    "params": {
                        "domain": "blacklanternsecurity.com",
                        "company": None,
                        "type": None,
                        "seniority": None,
                        "department": None,
                    },
                },
            },
        )
        module_test.httpx_mock.add_response(
            url="https://api.hunter.io/v2/domain-search?domain=blacklanternsecurity.com&api_key=4321&limit=100&offset=100",
            json={
                "data": {
                    "domain": "blacklanternsecurity.com",
                    "disposable": False,
                    "webmail": False,
                    "accept_all": False,
                    "pattern": "{first}",
                    "organization": "Black Lantern Security",
                    "description": None,
                    "twitter": None,
                    "facebook": None,
                    "linkedin": "https://linkedin.com/company/black-lantern-security",
                    "instagram": None,
                    "youtube": None,
                    "technologies": ["jekyll", "nginx"],
                    "country": "US",
                    "state": "CA",
                    "city": "Night City",
                    "postal_code": "12345",
                    "street": "123 Any St",
                    "emails": [
                        {
                            "value": "fdsa@blacklanternsecurity.com",
                            "type": "generic",
                            "confidence": 77,
                            "sources": [
                                {
                                    "domain": "blacklanternsecurity.com",
                                    "uri": "http://blacklanternsecurity.com",
                                    "extracted_on": "2021-06-09",
                                    "last_seen_on": "2023-03-21",
                                    "still_on_page": True,
                                }
                            ],
                            "first_name": None,
                            "last_name": None,
                            "position": None,
                            "seniority": None,
                            "department": "support",
                            "linkedin": None,
                            "twitter": None,
                            "phone_number": None,
                            "verification": {"date": None, "status": None},
                        }
                    ],
                    "linked_domains": [],
                },
                "meta": {
                    "results": 1,
                    "limit": 100,
                    "offset": 0,
                    "params": {
                        "domain": "blacklanternsecurity.com",
                        "company": None,
                        "type": None,
                        "seniority": None,
                        "department": None,
                    },
                },
            },
        )

    def check(self, module_test, events):
        assert any(e.data == "asdf@blacklanternsecurity.com" for e in events), "Failed to detect email #1"
        assert any(e.data == "fdsa@blacklanternsecurity.com" for e in events), "Failed to detect email #2"
